(* Semantic checking for Cryptal compiler *)

open Ast

module StringMap = Map.Make(String)

(* Semantic checking of a program. Returns void if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (globals, functions) =

  (* Raise an exception if the given list has a duplicate *)
  let report_duplicate exceptf list =
    let rec helper = function
	n1 :: n2 :: _ when n1 = n2 -> raise (Failure (exceptf n1))
      | _ :: t -> helper t
      | [] -> ()
    in helper (List.sort compare list)
  in

  (* Raise an exception if a given binding is to a void type *)
  let check_not_void exceptf = function
      (Void, n) -> raise (Failure (exceptf n))
    | _ -> ()
  in

  (* Raise an exception of the given rvalue type cannot be assigned to
     the given lvalue type *)
  let check_assign lvaluet rvaluet err =
     if lvaluet == rvaluet
     || lvaluet == Int && rvaluet == Gem
     || lvaluet == Int && rvaluet == Lattice
     || lvaluet == Lattice && rvaluet == Int
     || lvaluet == Lattice && rvaluet == String
     || lvaluet == Lattice && rvaluet == Gem
     || lvaluet == String && rvaluet == Char
     || lvaluet == Int && rvaluet == Char
     || lvaluet == Char && rvaluet == Int
     || lvaluet == Bool && rvaluet == Int
     then lvaluet else raise err
  in

  (**** Checking Global Variables ****)

  List.iter (check_not_void (fun n -> "illegal void global " ^ n)) globals;

  report_duplicate (fun n -> "duplicate global " ^ n) (List.map snd globals);

  (**** Checking Functions ****)

  if List.mem "print" (List.map (fun fd -> fd.fname) functions)
  then raise (Failure ("function print may not be defined")) else ();
  if List.mem "print_gem" (List.map (fun fd -> fd.fname) functions)
  then raise (Failure ("function print_gem may not be defined")) else ();
  if List.mem "print_lat" (List.map (fun fd -> fd.fname) functions)
  then raise (Failure ("function print_lat may not be defined")) else ();
  if List.mem "hash_md5" (List.map (fun fd -> fd.fname) functions)
  then raise (Failure ("function hash_md5 may not be defined")) else ();

  report_duplicate (fun n -> "duplicate function " ^ n)
    (List.map (fun fd -> fd.fname) functions);

  (* Function declaration for a named function *)
  let built_in_decls =
        StringMap.add "print"
        { typ = Void; fname = "print"; formals = [(Int, "x")];
       locals = []; body = [] }
       (StringMap.add "print_gem"
       { typ = Void; fname = "print_gem"; formals = [(Gem, "x")];
       locals = []; body =[] }
       (StringMap.add "hash_md5"
       { typ = String; fname = "hash_md5"; formals = [(String, "x")];
       locals = []; body =[] }
      (StringMap.singleton "print_lat"
       { typ = Void; fname = "print_lat"; formals = [(Lattice, "x")];
       locals = []; body = [] })))
  in

  let function_decls = List.fold_left (fun m fd -> StringMap.add fd.fname fd m)
                         built_in_decls functions
  in

  let function_decl s = try StringMap.find s function_decls
       with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = function_decl "main" in (* Ensure "main" is defined *)

  let check_function func =

    List.iter (check_not_void (fun n -> "illegal void formal " ^ n ^
      " in " ^ func.fname)) func.formals;

    report_duplicate (fun n -> "duplicate formal " ^ n ^ " in " ^ func.fname)
      (List.map snd func.formals);

    List.iter (check_not_void (fun n -> "illegal void local " ^ n ^
      " in " ^ func.fname)) func.locals;

    report_duplicate (fun n -> "duplicate local " ^ n ^ " in " ^ func.fname)
      (List.map snd func.locals);

    (* Type of each variable (global, formal, or local *)
    let symbols = List.fold_left (fun m (t, n) -> StringMap.add n t m)
      StringMap.empty (globals @ func.formals @ func.locals )
    in

    let type_of_identifier s tree =
      try StringMap.find s  tree
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    (* Return the type of an expression or throw an exception *)
    let rec expr tree = function
	Literal _ -> Int
      | BoolLit _ -> Bool
      | Ch _ -> Char
      | Null -> Void
      | String _ -> String
      | Gem (e1, e2) -> let v1 = expr tree e1 and v2 = expr tree e2 in
        (match (v1, v2) with
            (Int, Int) -> Gem
          | (Gem, Int) -> Gem
          | (Lattice, Int) -> Gem
          | (Int, Gem) -> Gem
          | (Lattice, Gem) -> Gem
          | (Int, Lattice) -> Gem
          | (Gem, Lattice) -> Gem
          | (Lattice, Lattice) -> Gem
          | _ -> raise (Failure ("illegal constructor for gem (" ^ string_of_typ v1
        ^ "," ^ string_of_typ v2 ^ ")"))
        )
      | Lat _ -> Lattice
      | Id s -> type_of_identifier s tree
      | Binop(e1, op, e2) as e -> let t1 = expr tree e1 and t2 = expr tree e2 in
  	    (match op with
          (* Rules for general C operators *)
          Add | Sub | Mult | Div | Power  when t1 = Int && t2 = Int -> Int
        | Power                           when t1 = Int && t2 = Gem -> Int
      	| Equal | Neq                     when t1 = t2 || t1 = Int && t2 = Char || t1 = Char && t2 = Int || t1 = Int && t2 = Bool || t1 = Bool && t2 = Int || t1 = String && t2 = Char || t1 = Char && t2 = String -> Bool
      	| Less | Leq | Greater | Geq      when t1 = t2 || t1 = Int && t2 = Char || t1 = Char && t2 = Int || t1 = Int && t2 = Bool || t1 = Bool && t2 = Int || t1 = String && t2 = Char || t1 = Char && t2 = String -> Bool
      	| And | Or                        when t1 = Bool && t2 = Bool -> Bool
        | Mod                             when t1 = Int -> Int
        (* Rules for gems *)
        | Add | Sub | Mult | Div          when t1 = Gem && t2 = Gem || t1 = Gem && t2 = Int || t1 = Int && t2 = Gem -> Gem
        | Equal | Neq                     when t1 = Gem && t2 = Int || t1 = Int && t2 = Gem -> Bool
        | Less | Leq | Greater | Geq      when t1 = Gem && t2 = Int || t1 = Int && t2 = Gem -> Bool
        | Power                           when t1 = Gem && t2 = Gem || t1 = Gem && t2 = Int || t1 = Gem && t2 = Lattice-> Gem
        (* Rules for lattices *)
        | Add | Sub | Mult | Div          when t1 = Lattice && t2 = Lattice || t1 = Int && t2 = Lattice || t1 = Lattice && t2 = Int-> Lattice
        | Equal | Neq                     when t1 = Lattice && t2 = Int || t1 = Int && t2 = Lattice -> Bool
        | Less | Leq | Greater | Geq      when t1 = Int && t2 = Lattice || t1 = Lattice && t2 = Int -> Bool
        | Power                           when t1 = Lattice && t2 = Lattice || t1 = Lattice && t2 = Int || t1 = Int && t2 = Lattice  -> Lattice
        (* Rules for gems and lattices *)
        | Add | Sub | Mult | Div          when t1 = Gem && t2 = Lattice || t1 = Lattice && t2 = Gem -> Gem
        | Equal | Neq                     when t1 = Gem && t2 = Lattice || t1 = Lattice && t2 = Gem || t1 = Lattice && t2 = String || t1 = String && t2 = Lattice -> Bool
        | Less | Leq | Greater | Geq      when t1 = Gem && t2 = Lattice || t1 = Lattice && t2 = Gem || t1 = Lattice && t2 = String || t1 = String && t2 = Lattice -> Bool
        | Power                           when t1 = Lattice && t2 = Gem -> Lattice
        | Mod                             when t1 = Lattice -> Lattice
            | _ -> raise (Failure ("illegal binary operator " ^
                    string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                    string_of_typ t2 ^ " in " ^ string_of_expr e))
              )
      | Unop(op, e) as ex -> let t = expr tree e in
	      (match op with
  	      Neg       when t = Int -> Int
        | Neg       when t = Gem -> Gem
        | Neg       when t = Lattice -> Lattice
  	    | Not       when t = Bool -> Bool
        | Not       when t = Gem -> Gem
        | _ -> raise (Failure ("illegal unary operator " ^ string_of_uop op ^
  	  		string_of_typ t ^ " in " ^ string_of_expr ex)))
      | Noexpr -> Void
      | Assign(var, e) as ex -> let lt = type_of_identifier var tree
                                and rt = expr tree e in
        check_assign lt rt (Failure ("illegal assignment " ^ string_of_typ lt ^
				     " = " ^ string_of_typ rt ^ " in " ^
				     string_of_expr ex))
      | Call(fname, actuals) as call -> let fd = function_decl fname in
        (* if fname - "printf" *)
        if fname <> "print" then
         if List.length actuals != List.length fd.formals then
           raise (Failure ("expecting " ^ string_of_int
             (List.length fd.formals) ^ " arguments in " ^ string_of_expr call))
         else
           List.iter2 (fun (ft, _) e -> let et = expr tree e in
              ignore (check_assign ft et
                (Failure ("illegal actual argument found " ^ string_of_typ et ^
                " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e))))
             fd.formals actuals;
           fd.typ
    in

    let check_num_expr tree e = if expr tree e != Int && expr tree e != Gem && expr tree e != Lattice
      then raise (Failure ("expected Number in " ^ string_of_expr e))
      else (true) in

    let check_bool_expr tree e = if expr tree e != Bool && check_num_expr tree e != true
     then raise (Failure ("expected Boolean expression in " ^ string_of_expr e))
     else () in


    (* Verify a statement or throw an exception *)
    let rec stmt tree = function
	   Block (_, sl) -> let rec check_block block_tree = function
         [ Return _ as s ] -> stmt block_tree s
         | Return _ :: _ -> raise (Failure "nothing may follow a return")
         | (Block (_,_) as b) :: ss ->stmt block_tree b; check_block block_tree ss
         | s :: ss -> stmt block_tree s ; check_block block_tree ss
         | [] -> ()
        in check_block tree sl
      | Expr e -> ignore (expr tree e)
      | Return e -> let t = expr tree e in if t = func.typ then () else
         raise (Failure ("return gives " ^ string_of_typ t ^ " expected " ^
                         string_of_typ func.typ ^ " in " ^ string_of_expr e))

      | If(p, b1, b2) -> check_bool_expr tree p; stmt tree b1; stmt tree b2
      | For(e1, e2, e3, st) -> ignore (expr tree e1); check_bool_expr tree e2;
                               ignore (expr tree e3); stmt tree st
      | While(p, s) -> check_bool_expr tree p; stmt tree s
    in

    let tree = List.fold_left (fun m (t,n) -> StringMap.add n t m)
      StringMap.empty (globals @ func.formals @ func.locals) in
    stmt tree (Block ([], func.body))

  in
  List.iter check_function functions
